function [mmNet,mmNetS,mmNetPattern,data,mzList,rule,ruleList,dataAll,mzName] = COVAIN_mzGroupAnalyzer(obj,evt)
% 

% ----------------------------------------------------------------------
% [mmNet,mmNetS,mmNetPattern,data,mzList,rule,ruleList,dataAll,mzName] =
% COVAIN_mzGroupAnalyzer(obj,evt);
% ----------------------------------------------------------------------
% Functions:  
%   1. read a list of mz values from single or multiple Excel files;  
%   2. combine duplicate peaks into one;
%   3. infer possible chemical group transformations between mz values
%      according to rules provided by users;
%   4. generate Pajek-compitable network file;
%   5. calculate the frequency of all possible one step transformations.
% ----------------------------------------------------------------------
% Input formats:
%   1. mz list file: first row, use "C, H, O" etc, for the composition
%   2. mz list file: second row is empty
%   3. rule file: first row, from the 3rd column, use "C, H, O" etc, for the 
%      net change, and "C H O" etc should be in the same order as the mz file
%   4. rule file: second row is not empty
% ----------------------------------------------------------------------
% Output explanation:
%   mmNet:        metabolite-metabolite network. connected: 1, not connected: 0.
%   mmNetS:       sparse matrix of mmNet, for further graph analysis
%   mmNetPattern: the values correspond to the rule numbers
%   data:         the composition of all (combined) peaks
%   mzList:       the value of (combined) peaks
%   rule:         the composition of the listed one-step transformation groups
%   ruleList:     the listed one-step transformation groups
%   dataAll:      the full data table
%   mzName:       the mz value names (string)
%   compReal:     the real composition elements (C/H/O/N, etc)
% ----------------------------------------------------------------------
% Xiaoliang Sun @ 2012-12-15 20:10
%               @ 2012-12-16 19:41
%               @ 2012-12-20 23:20                             
%               @ 2014-01-25, get dataAll which includes the time series information
%               @ 2014-02-04, integrate mzPathViewer
%               @ 2014-02-06, add other elements than C/H/O
%               @ 2014-06-11, support Mac to export results

%% user configurable lines
chemical_element_names = {'C','H','O','N','P','S','Na','K'};
number_of_exported_patterns = 1000;

%% get saving path
h = findobj('Tag','covainPanel_TOP');
dataStruct = guidata(h);

% if saving path exists (i.e., has loaded data)
if dataStruct.loadNumber > 0
    currentData = dataStruct.dataset(dataStruct.current).dataInfo;
    savePath = currentData.savePath;
else
    savePath = [];
end

%% mass definition
compNameVec = chemical_element_names;

%% Load mz data files, multiple selections are allowed
[dataFileName,dataFilePath] = uigetfile({'*.xlsx;*.xls','All Excel Files'},...
                            'MultiSelect','on','Select data file(s)');
if isnumeric(dataFileName) % if nothing is loaded, exit
    return;
end
if isempty(savePath)
    savePath = dataFilePath;
end

%% Load the rule file, single file
[ruleFileName, ruleFilePath] = uigetfile({'*.xlsx','*.xls'},...
                            'MultiSelect','off','Select the rule file');
if isnumeric(ruleFileName) % if nothing is loaded, exit
    warndlg('Error: no rule is loaded'); return;
end

%% determine the output file names
% 1. name the output file(s)
[fileName,pathName,flgStatus] = uiputfile([savePath,'*.xlsx'],...
                    'Please name the output files (single name)');
if flgStatus == 1
    outName = [pathName,'mzGroup_',fileName(1:max(findstr('.',fileName))-1)];
else
    return;
end

%% Read the m/z data as well as rules, and filter empty C/H/O/... rules
% Q: why not reading data when locating the m/z files??
% A: because a bug in Matlab 2012b, which will lose focus of the current
% window, and will not execute the uigetfile again. For other user's
% report, see:
% http://www.mathworks.com/matlabcentral/answers/58273
[rule,ruleList] = readRule(ruleFileName,ruleFilePath);
[data,compIdx,mzList,dataAll] = readData(dataFileName,dataFilePath);
compReal = compNameVec(compIdx);
if size(rule,2) < length(compIdx)
    warndlg('The rule file is not compitable with m/z data file! ');
    return;
end
rule = rule(:,compIdx);

% revised @ 2014-02-10, removing rules with all zeros
ruleList(all(rule==0,2),:) = [];
rule(all(rule==0,2),:) = [];


%% calculate 
% 1. the distance between compositions
hw = waitbar(0,'Calculating patterns ...');
dist = zeros(size(data,1)^2,size(data,2));
for i=1:size(data,1)
    for j=1:size(data,1)
        dist((i-1)*size(data,1)+j,:) = data(j,:) - data(i,:);
    end
end
% 2. change 3-bit dist into 1-bit
nTimes = fix(log10(max(abs(dist)))) + 1; 
nCumSum = cumsum(nTimes(1:end-1)); nCumSum = [nCumSum(end:-1:1),0];
distOne = sum(repmat(10.^nCumSum,size(dist,1),1).*dist,2);
ruleOne = sum(repmat(10.^nCumSum,size(rule,1),1).*rule,2);
% get the metabolite-metabolite network
distPattern = zeros(size(dist,1),1);
nHit = zeros(length(ruleOne),1);
for i=1:length(ruleOne)
    idxOne = (distOne == ruleOne(i));
    if ~isempty(idxOne)  % if hit
        nHit(i) = sum(idxOne);
        distPattern(idxOne) = i;
    end
end
mmNetPattern = reshape(distPattern,size(data,1),size(data,1))';  % ------ !!
mmNet = (mmNetPattern ~= 0);
mmNetS = sparse(1./mmNet);
% prepare the text output
outTable = cell(sum(mmNet(:)),7 + size(data,2));
outTable{1,1} = 'Precursor_From'; 
outTable{1,2} = 'TheoMass_From';
outTable{1,3} = 'Formula_From';
outTable{1,4} = 'Precursor_To'; 
outTable{1,5} = 'TheoMass_To';
outTable{1,6} = 'Formula_To';
outTable{1,7} = 'Transformation';
for i=1:size(data,2)
    outTable{1,i+7} = ['Diff ',compReal{i}];
end
waitbar(.20,hw,'Producing Pajek network file ...'); % updating waitbar
% produce the output table and Pajek format network, and mzName
% 1. simple one, no colors
mzName = cell(size(data,1),1);
fid = fopen([outName,'_Pajek.net'],'w');
if fid <= 0
    return;
end
fprintf(fid,['*Vertices    ',num2str(length(mzList)),'\n']);
for i=1:size(data,1)
    fprintf(fid,['    ',num2str(i),'  "mz',num2str(mzList(i)),'"\n']);
    mzName{i} = ['mz ',num2str(mzList(i))];
end
fprintf(fid,'*Arcs\n');
fprintf(fid,'*Edges\n');
countEdge = 0;
for i=1:size(mmNet,1)
    for j=1:size(mmNet,2)
        if mmNet(i,j)
            fprintf(fid,['    ',num2str(i),'  ',num2str(j),'  1\n']);
            countEdge = countEdge + 1;
            outTable{countEdge+1,1} = dataAll{i,1};
            outTable{countEdge+1,2} = mzList(i);
            outTable{countEdge+1,3} = dataAll{i,4};
            outTable{countEdge+1,4} = dataAll{j,1};
            outTable{countEdge+1,5} = mzList(j);
            outTable{countEdge+1,6} = dataAll{j,4};
            outTable{countEdge+1,7} = ruleList{mmNetPattern(i,j)};
            outTable(countEdge+1,8:end) = num2cell(rule(mmNetPattern(i,j),:));
        end
    end
end
fclose(fid);
waitbar(.40,hw,'Finding frequent patterns ...');
% preparing tables
freqTable0 = cell(length(ruleList)+1,size(data,2)+2);
freqTable0{1,1} = 'Count'; freqTable0{1,2} = 'Name';
freqTable0(1,3:end) = outTable(1,8:end);
freqTable0(2:end,1) = num2cell(nHit);
freqTable0(2:end,2) = ruleList;
freqTable0(2:end,3:end) = num2cell(rule);
% 2. colored vertices, such as: blue for <mz500 and red for others
                            %
                            %
                            %
% get statistics of distOne, other patterns
distOther = distOne(distPattern==0 & distOne ~= 0);
% questions on calculating the most frequent one-step transformation patterns 
% other than the listed ones
% export most frequent x_number of patterns !!
nFreq = min(number_of_exported_patterns,length(unique(distOther))); 

% calculate the most nFreq patterns
distOtherU = unique(distOther); lenU = length(distOtherU);
distOtherU_count = histc(distOther,distOtherU);
distOneFreq = [distOtherU,distOtherU_count]; % col 1: value; col 2: count
clear distOtherU*;  % to free memory
[~,idxFreq] = sort(distOneFreq(:,2),'descend');
distOneFreq = distOneFreq(idxFreq,:);   % all
distOneFreq = distOneFreq(1:nFreq,:);   % the first nFreq ones

% change the 1-bit dist to the original n-bit compositions
freqComp = zeros(nFreq, size(data,2));
weirdGroup = find(fix(10.^(1:100) .* 10.^(-1:-1:-100)) == 0);
for i=1:nFreq
    t = distOneFreq(i,1);
    for j=1:size(data,2)
        % freqComp(i,j) = fix(t*10^(-nCumSum(j)));
        % weird: fix(10^5 * 10^-5) = 0 !!
        if ismember(nCumSum(j),weirdGroup)
            t = t*10;
            freqComp(i,j) = fix(t*10^(-nCumSum(j)-1));
            t = t - freqComp(i,j)*10^(nCumSum(j)+1);
            t = t/10;
        else
            freqComp(i,j) = fix(t*10^(-nCumSum(j)));
            t = t - freqComp(i,j)*10^nCumSum(j);
        end
    end
end
% output the freq table
freqTable = cell(nFreq+1,size(data,2)+2);
freqTable{1,1} = 'Count'; freqTable{1,2} = 'Formular'; 
freqTable(1,3:end) = outTable(1,8:end);
freqTable(2:end,1) = num2cell(distOneFreq(:,2));
freqTable(2:end,3:end) = num2cell(freqComp);
for i=1:nFreq  % translate numbers to formular
    t = freqComp(i,:); s = [];
    if all(t(t~=0) < 0)
        t = - t; flgMinus = 1;
    else
        flgMinus = 0;
    end
    for j=1:length(compReal)
        if t(j) == 1
            s = [s,compReal{j}];
        elseif t(j) > 1
            s = [s,compReal{j},num2str(t(j))];
        end
    end
    if flgMinus == 1
        s = ['-',s];
    end
    freqTable{i+1,2} = s;
    waitbar(.40+.2*i/nFreq,hw,['Find the ',num2str(i),'-th pattern of all ',...
        num2str(lenU), ' patterns.']);
end

%% save all results to Excel
% if pc user ------ updated @ 2014-06-11 -------
if ispc==1
    waitbar(.60,hw,'Exporting transformation tables ...');
    xlswrite([outName,'_transformations.xlsx'],outTable);
    xlswrite([outName,'_freqListedGroups.xlsx'],freqTable0);
    xlswrite([outName,'_freqNotListedGroups.xlsx'],freqTable);
else
    waitbar(.60,hw,'Exporting transformation tables ...');
    fidString = cell(1,3);
    fidString{1} = [outName,'_transformations.txt'];
    fidString{2} = [outName,'_freqListedGroups.txt'];
    fidString{3} = [outName,'_freqNotListedGroups.txt'];
    outs = cell(1,3);
    outs{1} = outTable; outs{2} = freqTable0; outs{3} = freqTable;
    for m=1:length(fidString)
        fid_m = fopen(fidString{m},'w');
        outs_m = outs{m};
        if fid_m > 0
            for i=1:size(outs_m,1)
                for j=1:size(outs_m,2)
                    if isnumeric(outs_m{i,j}) % string output format
                        formatUnit = '%d';
                    else
                        formatUnit = '%s';
                    end
                    if j~=size(outs_m,2)
                        fprintf(fid_m,[formatUnit,'\t'],outs_m{i,j});
                    else
                        fprintf(fid_m,[formatUnit,'\n'],outs_m{i,j});
                    end
                end
            end
            fclose(fid_m); 
        else
            warning('Cannot export results. Please close file and try again.');
        end
    end
    clear outs; % remove this big guy
end
waitbar(.80,hw,'Exporting Matlab workspace ...');
save([outName,'.mat'],'mmNet','mmNetS','mmNetPattern','data','mzList','rule',...
    'ruleList','dataAll','mzName','compReal');
delete(hw);
answerNet = questdlg('View the network(fast) and do mudular analysis(slow)?',...
    'Network Analysis','No','View the network only','Do both','View the network only');
if strncmp(answerNet,'Vi',2)
    net = double((mmNet + mmNet')~=0);
    for i=1:length(net)
        net(i,i) = 0;
    end
    COVAIN_netView(net,mzName,'View',0);
elseif strncmp(answerNet,'Do',2);
    net = double((mmNet + mmNet')~=0);
    for i=1:length(net)
        net(i,i) = 0;
    end
    COVAIN_netView(net,mzName,'Modu',0);
else
    delete(answerNet);
end
    
%% get all pathways
% find if the toolboxes are available
toolboxList = arrayfun(@(x) x.Name, ver, 'UniformOutput',0);
bioinfoFlag = any(~cellfun(@isempty, regexp(toolboxList,'Bioinformatics Toolbox')));
parallelFlag = any(~cellfun(@isempty, regexp(toolboxList,'Parallel Computing Toolbox')));
if ~bioinfoFlag
    warndlg('Pathway searching needs Bioinfomatics Toolbox, which is not found in your Matlab.');
else
    answer = questdlg('Would you search for all pathways between m/z features?',...
        'Pathway searching','Yes','No','Load','Yes');
    if strcmp(answer,'Yes')
        if parallelFlag
            if size(data,1) >= 500
                msgbox('The dataset is big. Pathway searching may be time-consuming, but it can be accelerated by parallel computing.');
            end
            % pathName is the true saving path of mz output files, 
            % but savePath is COVAIN main save path !
            set(h,'HandleVisibility','on'); % handle visible
            COVAIN_mzGetPaths([outName,'.mat'],pathName); 
            set(h,'HandleVisibility','callback'); % handle not visible
        else
            if size(data,1) >= 500
                answer2 = questdlg('The dataset is big. Pathway searching may be time-consuming, but it can be accelerated by parallel computing. However, Parallel Computing Toolbox is not found in your Matlab. Please confirm you wish to proceed.','Confirmation to proceed','Yes','No','No');
                if strcmp(answer2,'No')
                    return;
                else
                    set(h,'HandleVisibility','on'); % handle visible
                    COVAIN_mzGetPaths([outName,'.mat'],pathName);
                    set(h,'HandleVisibility','callback'); % handle not visible
                end
            end
        end
    elseif strcmp(answer,'No') 
        return;
    else
        [mzFileName,mzFilePath] = uigetfile({'*.mat','mzStruct workspace'},...
                            'MultiSelect','off','Load the existing pathway dataset');
         if isnumeric(mzFileName) % if nothing is loaded, exit
             return;
         else 
             mzFullPath = [mzFilePath,mzFileName];
             pathStruct.dataType = 'external';
             pathStruct.dataPath = mzFullPath;
             pathStruct.preSavePath = pathName;
             set(findobj('Tag','button_mzViewer'),'UserData',pathStruct);
             COVAIN_mzPathViewer(findobj('Tag','button_mzViewer'));
         end
    end
end
    

%% ++++++++++++++  Sub function readData +++++++++++++++
    function [dataSub,compIdxSub,mzValue,rawData] = readData(dataFileName, dataFilePath)
        % called: [data,compIdx,mzList,dataAll] = readData(dataFileName,dataFilePath);
        % read the first file
        if iscell(dataFileName)
            name1 = dataFileName{1}; 
        else
            name1 = dataFileName; 
        end
        [~,~,raw] = xlsread([dataFilePath,name1]);
        % add data source
        raw = [raw, repmat(num2cell(1),size(raw,1),1)];
        raw{1,end} = 'Data source';  % change raw header
        % read the other files one by one, if any
        if iscell(dataFileName)
            hWait = waitbar(0,['Reading ',dataFileName{2}]);
            for k = 2:length(dataFileName) % add data source column, yes!
                if k > 2 % avoid duplicated waitbar (although no effect)
                    waitbar((k-1)/(length(dataFileName)-1),hWait,['Reading ',dataFileName{k}]);
                end
                [~,~,temp] = xlsread([dataFilePath,dataFileName{k}]);
                % discard headers and add source
                raw = [raw; temp(3:end,:),repmat(num2cell(k),size(temp,1)-2,1)]; 
            end
            close(hWait);
        end
        
%         % clean redundunt rows and columns ----- a special case
%         % useless when adding data source
%         try
%             flgRowEmpty = all(cellfun(@isnan,raw(2,:))==1);
%             if flgRowEmpty
%                 raw(2,:) = []; 
%             end
%         catch  %#ok<CTCH>
%             % do nothing
%         end
        raw(2,:) = [];
        raw(cellfun(@isempty,raw(2:end,1)),:) = []; % if mz value is empty
        
        % ****** @ 2014-02-07 ******
        % remove rows with empty composition 
        idxColComp = strmatch('Composition',raw(1,:),'exact');
        tempVar = cellfun(@(x) isnan(x), raw(:,idxColComp),'UniformOutput',0);
        raw(cellfun(@(x) all(x), tempVar),:) = [];
        
        % backup table with combined name, intensity and ppm
        rawData = cell(size(raw,1)-1,5); % combined name, intensity, ppm, formula, source,
        colMZ = raw(2:end,strmatch('m/z',raw(1,:),'exact'));
        colTheoMass = raw(2:end, strmatch('Theo. Mass',raw(1,:),'exact'));
        colIntensity = raw(2:end, strmatch('Intensity',raw(1,:),'exact'));
        colPPM = raw(2:end, strmatch('Delta (ppm)',raw(1,:),'exact'));
        % colComposition = strtrim(raw(2:end,strmatch('Composition',raw(1,:),'exact')));
        colComposition = raw(2:end,strmatch('Composition',raw(1,:),'exact'));
        
        % fill values to such case:
        % Row1: A 10
        % Row2: A 
        % Row3: A
        % Row4: B 20
        % ...
        k = 1;
        while k <= size(colMZ,1) - 1
            count = 1;
            while k + count <= size(colMZ,1) && isnan(colMZ{k+count})
                colMZ(k+count) = colMZ(k);
                % colIntensity(k+count) = colIntensity(k); only mz is empty
                count = count + 1;
            end
            k = k + count;
        end
        raw(2:end,strmatch('m/z',raw(1,:),'exact')) = colMZ;
        
        % note the precision = %10.5f !!!
        for k = 1:size(raw,1)-1
            %rawData{k,1} = [num2str(colMZ{k},'%10.5f'),'_',num2str(colTheoMass{k},'%10.5f')];
            %rawData{k,1} = num2str(colTheoMass{k},'%10.5f');
            rawData{k,1} = num2str(colMZ{k},'%10.5f');
        end
        rawData(:,2) = colIntensity; 
        rawData(:,3) = colPPM; 
        rawData(:,4) = colComposition;
        rawData(:,5) = raw(2:end,end);  % data source, like [1], meaning "day 1"
        
        % ***** @ 2014-02-07 *****
        % full cells with empty theo.mass
        idxColTheo = strmatch('Theo. Mass',raw(1,:),'exact');
        tempVar = cellfun(@(x) isnan(x), raw(:,idxColTheo),'UniformOutput',0);
        raw(cellfun(@(x) all(x), tempVar),idxColTheo) = colMZ(cellfun(@(x) all(x), tempVar));       
         
        % ==========  This is the new 2014 part  =============
        % sort mz values. 
        % Note: same Theo. Mass m/z in different days is regarded as the same compound 
        % Note: but all days of this m/z feature are counted in 
        
        % ***** revised @ 2014-02-07 ***** theo mass not unique
        % combine colTheoMass and colComposition
        strMassComp = cell(length(colTheoMass),1);
        numBits = floor(log10(cell2mat(colTheoMass)));
        gapBits = max(numBits) - numBits;
        for k=1:length(colTheoMass)
            massStr = num2str(colTheoMass{k});
            massStr = [repmat('0',1,gapBits(k)),massStr]; % add 0 before short numbers
            strMassComp{k} = [massStr,'---',colComposition{k}];
        end
        [mzU,idxSub] = unique(strMassComp,'first');
        
        % update rawData
        rawData = rawData(idxSub,:);
        % count all days in
        for k = 1:length(mzU)
            idx_mzU = sort(find(  strcmp(mzU{k},strMassComp)  ));
            rawData{k,5} = cell2mat(raw(idx_mzU+1,end))'; % transpose to row
        end
        % update raw
        raw = [raw(1,:); raw(idxSub+1,:)];
        
        % ======================================================= 
               
        % read C, H, O, etc
        numComp = zeros(size(raw,1)-1,length(compNameVec));
        abcComp = raw(2:end,strmatch('Composition',raw(1,:),'exact'));
        hComp = waitbar(0,'Reading C/H/O,etc compositions ... ');
        for k1 = 1:size(numComp,1)
            waitbar(k1/size(numComp,1));
            for k2 = 1:size(numComp,2)
                if ~isempty(strfind(abcComp{k1},compNameVec{k2}))
                    temp = regexp(abcComp{k1},['(?<=',compNameVec{k2},')[0-9]+'],'match');
                    if isempty(temp)
                        numComp(k1,k2) = 1;
                    else
                        numComp(k1,k2) = str2double(cell2mat(temp));
                    end
                end
            end
        end
        close(hComp);
        compIdxSub = find(any(numComp)); % remove all zero columns
        numComp = numComp(:,compIdxSub);
        
        % get output
        dataSub = numComp;
        mzValue = cell2mat(raw(2:end,strmatch('Theo. Mass',raw(1,:),'exact')));
    end                 
        
%% ++++++++++++++  Sub function readRule +++++++++++++++ 
    function [ruleSub,ruleListSub] = readRule(ruleFileName, ruleFilePath)
        [~,~,raw] = xlsread([ruleFilePath,ruleFileName]);
        % read rule list
        ruleListSub = raw(2:end,1);
        % read rule data
        ruleSub = cell2mat(raw(2:end,3:end));
    end

end  % end of main function



        
% % ======= formats of Pajek ========        
% *Vertices 9
% 1 "a" ic Pink bc Black
% 2 "b" ic CMYK(0,0,1,0.0) bc CMYKFF000000
% 3 "c" ic Cyan bc Yellow
% 4 "d" ic Purple bc Orange
% 5 "e" ic Orange bc Brown
% 6 "f" ic Magenta bc Green
% 7 "g" ic Brown bc Magenta
% 8 "h" ic RGB(1,0,0) bc Blue
% 9 "i" ic Green bc Magenta
% *Arcs
% 1 2 1 c RGB0000FF
% 2 3 1 c Red
% 3 4 1 c Black
% 4 5 1 c Yellow
% 5 6 1 c Gray
% 6 7 1 c Cyan
% 7 8 1 c Magenta
% 8 9 1 c Purple
% 9 1 1 c Brown        
        
