function mzGroupAnalyzer
% No GUI version of COVAIN_mzGroupAnalyzer & COVAIN_mzGetPaths
% The output results can be viewed by COVAIN_mzPathViewer
% Xiaoliang Sun @ 2014-06-19

% ----------------------------------------------------------------------
% [mmNet,mmNetS,mmNetPattern,data,mzList,rule,ruleList,dataAll,mzName] =
% COVAIN_mzGroupAnalyzer(obj,evt);
% ----------------------------------------------------------------------
% Functions:
%   1. read a list of mz values from single or multiple Excel files;
%   2. combine duplicate peaks into one;
%   3. infer possible chemical group transformations between mz values
%      according to rules provided by users;
%   4. generate Pajek-compitable network file;
%   5. calculate the frequency of all possible one step transformations.
% ----------------------------------------------------------------------
% Input formats:
%   1. mz list file: first row, use "C, H, O" etc, for the composition
%   2. mz list file: second row is empty
%   3. rule file: first row, from the 3rd column, use "C, H, O" etc, for the
%      net change, and "C H O" etc should be in the same order as the mz file
%   4. rule file: second row is not empty
% ----------------------------------------------------------------------
% Output explanation:
%   mmNet:        metabolite-metabolite network. connected: 1, not connected: 0.
%   mmNetS:       sparse matrix of mmNet, for further graph analysis
%   mmNetPattern: the values correspond to the rule numbers
%   data:         the composition of all (combined) peaks
%   mzList:       the value of (combined) peaks
%   rule:         the composition of the listed one-step transformation groups
%   ruleList:     the listed one-step transformation groups
%   dataAll:      the full data table
%   mzName:       the mz value names (string)
%   compReal:     the real composition elements (C/H/O/N, etc)
% ----------------------------------------------------------------------
% Xiaoliang Sun @ 2012-12-15 20:10
%               @ 2012-12-16 19:41
%               @ 2012-12-20 23:20
%               @ 2014-01-25, get dataAll which includes the time series information
%               @ 2014-02-04, integrate mzPathViewer
%               @ 2014-02-06, add other elements than C/H/O
%               @ 2014-06-11, support Mac to export results

%% user configurable lines
chemical_element_names = {'C','H','O','N','P','S','Na','K'};
number_of_exported_patterns = 1000;

%% mass definition
compNameVec = chemical_element_names;

%% Load mz data files, multiple selections are allowed
[dataFileName,dataFilePath] = uigetfile({'*.xlsx;*.xls','All Excel Files'},...
    'MultiSelect','on','Select data file(s)');
if isnumeric(dataFileName) % if nothing is loaded, exit
    return;
end

%% Load the rule file, single file
[ruleFileName, ruleFilePath] = uigetfile({'*.xlsx','*.xls'},...
    'MultiSelect','off','Select the rule file');
if isnumeric(ruleFileName) % if nothing is loaded, exit
    warndlg('Error: no rule is loaded'); return;
end

%% determine the output file names
% 1. name the output file(s)
savePath = uigetdir(dataFilePath,'Please choose a folder to save results');
if savePath == 0   %
    savePath = dataFilePath;
else
    savePath = [savePath,'\'];
end
[fileName,pathName,flgStatus] = uiputfile([savePath,'*.xlsx'],...
    'Please name the output files (single name)');
if flgStatus == 1
    outName = [pathName,'mzGroup_',fileName(1:max(findstr('.',fileName))-1)];
else
    return;
end

%% Read the m/z data as well as rules, and filter empty C/H/O/... rules
% Q: why not reading data when locating the m/z files??
% A: because a bug in Matlab 2012b, which will lose focus of the current
% window, and will not execute the uigetfile again. For other user's
% report, see:
% http://www.mathworks.com/matlabcentral/answers/58273
[rule,ruleList] = readRule(ruleFileName,ruleFilePath);
[data,compIdx,mzList,dataAll] = readData(dataFileName,dataFilePath);
compReal = compNameVec(compIdx);
if size(rule,2) < length(compIdx)
    warndlg('The rule file is not compitable with m/z data file! ');
    return;
end
rule = rule(:,compIdx);

% revised @ 2014-02-10, removing rules with all zeros
ruleList(all(rule==0,2),:) = [];
rule(all(rule==0,2),:) = [];


%% calculate
% 1. the distance between compositions
hw = waitbar(0,'Calculating patterns ...');
dist = zeros(size(data,1)^2,size(data,2));
for i=1:size(data,1)
    for j=1:size(data,1)
        dist((i-1)*size(data,1)+j,:) = data(j,:) - data(i,:);
    end
end
% 2. change 3-bit dist into 1-bit
nTimes = fix(log10(max(abs(dist)))) + 1;
nCumSum = cumsum(nTimes(1:end-1)); nCumSum = [nCumSum(end:-1:1),0];
distOne = sum(repmat(10.^nCumSum,size(dist,1),1).*dist,2);
ruleOne = sum(repmat(10.^nCumSum,size(rule,1),1).*rule,2);
% get the metabolite-metabolite network
distPattern = zeros(size(dist,1),1);
nHit = zeros(length(ruleOne),1);
for i=1:length(ruleOne)
    idxOne = (distOne == ruleOne(i));
    if ~isempty(idxOne)  % if hit
        nHit(i) = sum(idxOne);
        distPattern(idxOne) = i;
    end
end
mmNetPattern = reshape(distPattern,size(data,1),size(data,1))';  % ------ !!
mmNet = (mmNetPattern ~= 0);
mmNetS = sparse(1./mmNet);
% prepare the text output
outTable = cell(sum(mmNet(:)),7 + size(data,2));
outTable{1,1} = 'Precursor_From';
outTable{1,2} = 'TheoMass_From';
outTable{1,3} = 'Formula_From';
outTable{1,4} = 'Precursor_To';
outTable{1,5} = 'TheoMass_To';
outTable{1,6} = 'Formula_To';
outTable{1,7} = 'Transformation';
for i=1:size(data,2)
    outTable{1,i+7} = ['Diff ',compReal{i}];
end
waitbar(.20,hw,'Producing Pajek network file ...'); % updating waitbar
% produce the output table and Pajek format network, and mzName
% 1. simple one, no colors
mzName = cell(size(data,1),1);
fid = fopen([outName,'_Pajek.net'],'w');
if fid <= 0
    return;
end
fprintf(fid,['*Vertices    ',num2str(length(mzList)),'\n']);
for i=1:size(data,1)
    fprintf(fid,['    ',num2str(i),'  "mz',num2str(mzList(i)),'"\n']);
    mzName{i} = ['mz ',num2str(mzList(i))];
end
fprintf(fid,'*Arcs\n');
fprintf(fid,'*Edges\n');
countEdge = 0;
for i=1:size(mmNet,1)
    for j=1:size(mmNet,2)
        if mmNet(i,j)
            fprintf(fid,['    ',num2str(i),'  ',num2str(j),'  1\n']);
            countEdge = countEdge + 1;
            outTable{countEdge+1,1} = dataAll{i,1};
            outTable{countEdge+1,2} = mzList(i);
            outTable{countEdge+1,3} = dataAll{i,4};
            outTable{countEdge+1,4} = dataAll{j,1};
            outTable{countEdge+1,5} = mzList(j);
            outTable{countEdge+1,6} = dataAll{j,4};
            outTable{countEdge+1,7} = ruleList{mmNetPattern(i,j)};
            outTable(countEdge+1,8:end) = num2cell(rule(mmNetPattern(i,j),:));
        end
    end
end
fclose(fid);
waitbar(.40,hw,'Finding frequent patterns ...');
% preparing tables
freqTable0 = cell(length(ruleList)+1,size(data,2)+2);
freqTable0{1,1} = 'Count'; freqTable0{1,2} = 'Name';
freqTable0(1,3:end) = outTable(1,8:end);
freqTable0(2:end,1) = num2cell(nHit);
freqTable0(2:end,2) = ruleList;
freqTable0(2:end,3:end) = num2cell(rule);
% 2. colored vertices, such as: blue for <mz500 and red for others
%
%
%
% get statistics of distOne, other patterns
distOther = distOne(distPattern==0 & distOne ~= 0);
% questions on calculating the most frequent one-step transformation patterns
% other than the listed ones
% export most frequent x_number of patterns !!
nFreq = min(number_of_exported_patterns,length(unique(distOther)));

% calculate the most nFreq patterns
distOtherU = unique(distOther); lenU = length(distOtherU);
distOtherU_count = histc(distOther,distOtherU);
distOneFreq = [distOtherU,distOtherU_count]; % col 1: value; col 2: count
clear distOtherU*;  % to free memory
[~,idxFreq] = sort(distOneFreq(:,2),'descend');
distOneFreq = distOneFreq(idxFreq,:);   % all
distOneFreq = distOneFreq(1:nFreq,:);   % the first nFreq ones

% change the 1-bit dist to the original n-bit compositions
freqComp = zeros(nFreq, size(data,2));
weirdGroup = find(fix(10.^(1:100) .* 10.^(-1:-1:-100)) == 0);
for i=1:nFreq
    t = distOneFreq(i,1);
    for j=1:size(data,2)
        % freqComp(i,j) = fix(t*10^(-nCumSum(j)));
        % weird: fix(10^5 * 10^-5) = 0 !!
        if ismember(nCumSum(j),weirdGroup)
            t = t*10;
            freqComp(i,j) = fix(t*10^(-nCumSum(j)-1));
            t = t - freqComp(i,j)*10^(nCumSum(j)+1);
            t = t/10;
        else
            freqComp(i,j) = fix(t*10^(-nCumSum(j)));
            t = t - freqComp(i,j)*10^nCumSum(j);
        end
    end
end
% output the freq table
freqTable = cell(nFreq+1,size(data,2)+2);
freqTable{1,1} = 'Count'; freqTable{1,2} = 'Formular';
freqTable(1,3:end) = outTable(1,8:end);
freqTable(2:end,1) = num2cell(distOneFreq(:,2));
freqTable(2:end,3:end) = num2cell(freqComp);
for i=1:nFreq  % translate numbers to formular
    t = freqComp(i,:); s = [];
    if all(t(t~=0) < 0)
        t = - t; flgMinus = 1;
    else
        flgMinus = 0;
    end
    for j=1:length(compReal)
        if t(j) == 1
            s = [s,compReal{j}];
        elseif t(j) > 1
            s = [s,compReal{j},num2str(t(j))];
        end
    end
    if flgMinus == 1
        s = ['-',s];
    end
    freqTable{i+1,2} = s;
    waitbar(.40+.2*i/nFreq,hw,['Find the ',num2str(i),'-th pattern of all ',...
        num2str(lenU), ' patterns.']);
end

%% save all results to Excel
% if pc user ------ updated @ 2014-06-11 -------
if ispc==1
    waitbar(.60,hw,'Exporting transformation tables ...');
    xlswrite([outName,'_transformations.xlsx'],outTable);
    xlswrite([outName,'_freqListedGroups.xlsx'],freqTable0);
    xlswrite([outName,'_freqNotListedGroups.xlsx'],freqTable);
else
    waitbar(.60,hw,'Exporting transformation tables ...');
    fidString = cell(1,3);
    fidString{1} = [outName,'_transformations.txt'];
    fidString{2} = [outName,'_freqListedGroups.txt'];
    fidString{3} = [outName,'_freqNotListedGroups.txt'];
    outs = cell(1,3);
    outs{1} = outTable; outs{2} = freqTable0; outs{3} = freqTable;
    for m=1:length(fidString)
        fid_m = fopen(fidString{m},'w');
        outs_m = outs{m};
        if fid_m > 0
            for i=1:size(outs_m,1)
                for j=1:size(outs_m,2)
                    if isnumeric(outs_m{i,j}) % string output format
                        formatUnit = '%d';
                    else
                        formatUnit = '%s';
                    end
                    if j~=size(outs_m,2)
                        fprintf(fid_m,[formatUnit,'\t'],outs_m{i,j});
                    else
                        fprintf(fid_m,[formatUnit,'\n'],outs_m{i,j});
                    end
                end
            end
            fclose(fid_m);
        else
            warning('Cannot export results. Please close file and try again.');
        end
    end
    clear outs; % remove this big guy
end
waitbar(.80,hw,'Exporting Matlab workspace ...');
save([outName,'.mat'],'mmNet','mmNetS','mmNetPattern','data','mzList','rule',...
    'ruleList','dataAll','mzName','compReal');
delete(hw);

%% get all pathways
% find if the toolboxes are available
toolboxList = arrayfun(@(x) x.Name, ver, 'UniformOutput',0);
bioinfoFlag = any(~cellfun(@isempty, regexp(toolboxList,'Bioinformatics Toolbox')));
parallelFlag = any(~cellfun(@isempty, regexp(toolboxList,'Parallel Computing Toolbox')));
if ~bioinfoFlag
    warndlg('Pathway searching needs Bioinfomatics Toolbox, which is not found in your Matlab.');
else
    answer = questdlg('Would you search for all pathways between m/z features?',...
        'Pathway searching','Yes','No','Yes');
    if strcmp(answer,'Yes')
        if parallelFlag
            if size(data,1) >= 500
                msgbox('The dataset is big. Pathway searching may be time-consuming, but it can be accelerated by parallel computing.');
            end
            mzGetPaths;
        else
            if size(data,1) >= 500
                answer2 = questdlg('The dataset is big. Pathway searching may be time-consuming, but it can be accelerated by parallel computing. However, Parallel Computing Toolbox is not found in your Matlab. Please confirm you wish to proceed.','Confirmation to proceed','Yes','No','No');
                if strcmp(answer2,'No')
                    return;
                else
                    mzGetPaths;
                end
            end
        end
    elseif strcmp(answer,'No')
        return;
    end
end


%% ++++++++++++++  Sub function readData +++++++++++++++
    function [dataSub,compIdxSub,mzValue,rawData] = readData(dataFileName, dataFilePath)
        % called: [data,compIdx,mzList,dataAll] = readData(dataFileName,dataFilePath);
        % read the first file
        if iscell(dataFileName)
            name1 = dataFileName{1};
        else
            name1 = dataFileName;
        end
        [~,~,raw] = xlsread([dataFilePath,name1]);
        % add data source
        raw = [raw, repmat(num2cell(1),size(raw,1),1)];
        raw{1,end} = 'Data source';  % change raw header
        % read the other files one by one, if any
        if iscell(dataFileName)
            hWait = waitbar(0,['Reading ',dataFileName{2}]);
            for k = 2:length(dataFileName) % add data source column, yes!
                if k > 2 % avoid duplicated waitbar (although no effect)
                    waitbar((k-1)/(length(dataFileName)-1),hWait,['Reading ',dataFileName{k}]);
                end
                [~,~,temp] = xlsread([dataFilePath,dataFileName{k}]);
                % discard headers and add source
                raw = [raw; temp(3:end,:),repmat(num2cell(k),size(temp,1)-2,1)];
            end
            close(hWait);
        end
        
        %         % clean redundunt rows and columns ----- a special case
        %         % useless when adding data source
        %         try
        %             flgRowEmpty = all(cellfun(@isnan,raw(2,:))==1);
        %             if flgRowEmpty
        %                 raw(2,:) = [];
        %             end
        %         catch  %#ok<CTCH>
        %             % do nothing
        %         end
        raw(2,:) = [];
        raw(cellfun(@isempty,raw(2:end,1)),:) = []; % if mz value is empty
        
        % ****** @ 2014-02-07 ******
        % remove rows with empty composition
        idxColComp = strmatch('Composition',raw(1,:),'exact');
        tempVar = cellfun(@(x) isnan(x), raw(:,idxColComp),'UniformOutput',0);
        raw(cellfun(@(x) all(x), tempVar),:) = [];
        
        % backup table with combined name, intensity and ppm
        rawData = cell(size(raw,1)-1,5); % combined name, intensity, ppm, formula, source,
        colMZ = raw(2:end,strmatch('m/z',raw(1,:),'exact'));
        colTheoMass = raw(2:end, strmatch('Theo. Mass',raw(1,:),'exact'));
        colIntensity = raw(2:end, strmatch('Intensity',raw(1,:),'exact'));
        colPPM = raw(2:end, strmatch('Delta (ppm)',raw(1,:),'exact'));
        % colComposition = strtrim(raw(2:end,strmatch('Composition',raw(1,:),'exact')));
        colComposition = raw(2:end,strmatch('Composition',raw(1,:),'exact'));
        
        % fill values to such case:
        % Row1: A 10
        % Row2: A
        % Row3: A
        % Row4: B 20
        % ...
        k = 1;
        while k <= size(colMZ,1) - 1
            count = 1;
            while k + count <= size(colMZ,1) && isnan(colMZ{k+count})
                colMZ(k+count) = colMZ(k);
                % colIntensity(k+count) = colIntensity(k); only mz is empty
                count = count + 1;
            end
            k = k + count;
        end
        raw(2:end,strmatch('m/z',raw(1,:),'exact')) = colMZ;
        
        % note the precision = %10.5f !!!
        for k = 1:size(raw,1)-1
            %rawData{k,1} = [num2str(colMZ{k},'%10.5f'),'_',num2str(colTheoMass{k},'%10.5f')];
            %rawData{k,1} = num2str(colTheoMass{k},'%10.5f');
            rawData{k,1} = num2str(colMZ{k},'%10.5f');
        end
        rawData(:,2) = colIntensity;
        rawData(:,3) = colPPM;
        rawData(:,4) = colComposition;
        rawData(:,5) = raw(2:end,end);  % data source, like [1], meaning "day 1"
        
        % ***** @ 2014-02-07 *****
        % full cells with empty theo.mass
        idxColTheo = strmatch('Theo. Mass',raw(1,:),'exact');
        tempVar = cellfun(@(x) isnan(x), raw(:,idxColTheo),'UniformOutput',0);
        raw(cellfun(@(x) all(x), tempVar),idxColTheo) = colMZ(cellfun(@(x) all(x), tempVar));
        
        % ==========  This is the new 2014 part  =============
        % sort mz values.
        % Note: same Theo. Mass m/z in different days is regarded as the same compound
        % Note: but all days of this m/z feature are counted in
        
        % ***** revised @ 2014-02-07 ***** theo mass not unique
        % combine colTheoMass and colComposition
        strMassComp = cell(length(colTheoMass),1);
        numBits = floor(log10(cell2mat(colTheoMass)));
        gapBits = max(numBits) - numBits;
        for k=1:length(colTheoMass)
            massStr = num2str(colTheoMass{k});
            massStr = [repmat('0',1,gapBits(k)),massStr]; % add 0 before short numbers
            strMassComp{k} = [massStr,'---',colComposition{k}];
        end
        [mzU,idxSub] = unique(strMassComp,'first');
        
        % update rawData
        rawData = rawData(idxSub,:);
        % count all days in
        for k = 1:length(mzU)
            idx_mzU = sort(find(  strcmp(mzU{k},strMassComp)  ));
            rawData{k,5} = cell2mat(raw(idx_mzU+1,end))'; % transpose to row
        end
        % update raw
        raw = [raw(1,:); raw(idxSub+1,:)];
        
        % =======================================================
        
        % read C, H, O, etc
        numComp = zeros(size(raw,1)-1,length(compNameVec));
        abcComp = raw(2:end,strmatch('Composition',raw(1,:),'exact'));
        hComp = waitbar(0,'Reading C/H/O,etc compositions ... ');
        for k1 = 1:size(numComp,1)
            waitbar(k1/size(numComp,1));
            for k2 = 1:size(numComp,2)
                if ~isempty(strfind(abcComp{k1},compNameVec{k2}))
                    temp = regexp(abcComp{k1},['(?<=',compNameVec{k2},')[0-9]+'],'match');
                    if isempty(temp)
                        numComp(k1,k2) = 1;
                    else
                        numComp(k1,k2) = str2double(cell2mat(temp));
                    end
                end
            end
        end
        close(hComp);
        compIdxSub = find(any(numComp)); % remove all zero columns
        numComp = numComp(:,compIdxSub);
        
        % get output
        dataSub = numComp;
        mzValue = cell2mat(raw(2:end,strmatch('Theo. Mass',raw(1,:),'exact')));
    end

%% ++++++++++++++  Sub function readRule +++++++++++++++
    function [ruleSub,ruleListSub] = readRule(ruleFileName, ruleFilePath)
        [~,~,raw] = xlsread([ruleFilePath,ruleFileName]);
        % read rule list
        ruleListSub = raw(2:end,1);
        % read rule data
        ruleSub = cell2mat(raw(2:end,3:end));
    end


%% subfunction mzGetPaths
    function mzGetPaths
        
        len = length(mmNet); len2 = len*(len-1)/2;
        allShortPath = cell(len*(len-1)/2,1); % upper tri matrix - diag
        
        % get the indices
        x = zeros(len*(len-1)/2,1);
        y = zeros(len*(len-1)/2,1);
        ct = 0;
        for i1=1:len
            for j1=i1+1:len                    % upper tri !!! always y > x !!!
                ct = ct + 1;
                x(ct) = i1;  y(ct) = j1;
            end
        end
        
        % % set four parallel jobs only if the user has parallel computing toolbox
        if parallelFlag
            matlabpool;
            parfor i1 = 1:len2
                [~,allShortPath{i1}] = graphshortestpath(mmNetS,x(i1),y(i1));
            end
        else
            for i1 = 1:len2
                [~,allShortPath{i1}] = graphshortestpath(mmNetS,x(i1),y(i1));
            end
        end
        % save allShortPath0 allShortPath;
        
        % solid all paths
        idxAll = 1:len2;
        idxSolid = find(~cellfun(@isempty, allShortPath));
        x_solid = x(idxSolid);
        y_solid = y(idxSolid);
        allSolidPath = allShortPath(idxSolid);
        
        % filter out wrong time series
        allTimedPath = cell(length(idxSolid),1);
        flgCorrect = zeros(length(idxSolid),1);
        for k = 1:length(idxSolid)
            pathArray = dataAll(allSolidPath{k},5); % cell array containing time series seq
            nRow = size(pathArray,1);
            if nRow == 1
                thePath = pathArray{1};
                continue;
            end
            i1 = 1; thePath = [];
            while i1 <= nRow - 1
                row1 = pathArray{i1};
                row2 = pathArray{i1+1};
                if min(row2) >= min(row1)  % continue to next loop
                    thePath = [thePath, min(row1)];
                    i1 = i1 + 1;
                elseif max(row2) < min(row1)
                    thePath = -1;
                    break;             % fails
                else
                    nth = find(row2 >= min(row1), 1 ); % row2 values may not be consecutive
                    pathArray{i1+1} = pathArray{i1+1}(nth:end);
                    thePath = [thePath, min(row1)]; % continue, but remove smaller values in next row
                    i1 = i1 + 1;
                end
            end
            if thePath ~= -1
                thePath = [thePath,min(pathArray{i1})];
                allTimedPath{k} = thePath;
                flgCorrect(k) = 1;
            end
        end
        % update the indices
        allTimedPath = allTimedPath(flgCorrect==1);
        allCorrectPath = allSolidPath(flgCorrect==1);
        idxCorrect = idxSolid(flgCorrect==1);
        x_correct = x_solid(flgCorrect==1);
        y_correct = y_solid(flgCorrect==1);
        
        % elimate the correct paths
        fx = ones(length(x_correct),1);
        fy = ones(length(y_correct),1);
        siz = cellfun(@length, allCorrectPath);
        ct = max(siz);
        while ct > 1
            idx = find(siz == ct);   tic; % decreasing from longest paths
            % make it parallel in the level of i
            if ~isempty(idx)              % ct may not be consecutive
                idxElim = cell(length(idx),1);
                parfor i1=1:length(idx)
                    s1 = allCorrectPath{idx(i1)};
                    groupS = nchoosek(s1,2);     % find all intermediate paths
                    groupS(groupS(:,1)==s1(1) & groupS(:,2)==s1(end),:) = []; % except this one
                    % compare with allCorrectPath to judge if to elimate
                    for j1=1:size(groupS,1)
                        node1 = min(groupS(j1,:)); node2 = max(groupS(j1,:));
                        idxGlobal = find(x_correct == node1 & y_correct == node2);
                        if fx(idxGlobal) & fy(idxGlobal)
                            if groupS(j1,1) < groupS(j1,2)  % remember the upper tri !!!
                                sPath = s1(find(s1==groupS(j1,1)):find(s1==groupS(j1,2)));
                            else
                                sPath = s1(find(s1==groupS(j1,2)):-1:find(s1==groupS(j1,1)));
                            end
                            % get the shortest path from original allCorrectPath
                            tPath = allCorrectPath{idxGlobal};
                            if length(tPath) == length(sPath) && all(tPath == sPath)
                                idxElim{i1} = [idxElim{i1}; idxGlobal];
                            end
                        end
                    end
                end
                idxElimAll = [];
                for i1=1:length(idx)
                    idxElimAll = [idxElimAll; idxElim{i1}];
                end
                fx(idxElimAll) = 0;  fy(idxElimAll) = 0;
            end
            ct = ct - 1;
            disp([num2str(ct+1),' ...... ',num2str(toc),' ..... ',num2str(sum(fx))]);
        end
        if parallelFlag
            matlabpool close;
        end
        % save results
        idxFinal = find(fx == 1);
        finalCorrectPath = allCorrectPath(idxFinal);
        finalTimedPath = allTimedPath(idxFinal);
        x_final = x_correct(idxFinal);
        y_final = y_correct(idxFinal);
        
        
        % formulate allPath (all-in-one structure)
        % columns: mzFrom/choFrom/mzTo/choTo/rules/dispPath/mzPath/choPath/TS
        % % discarded: allPath = cell(length(finalCorrectPath),9);
        str0.mzFrom = []; str0.choFrom = []; str0.mzTo =[]; str0.choTo = []; str0.rules = [];
        str0.rulePath = []; str0.dispPath = []; str0.mzPath = []; str0.choPath = []; str0.TS = [];
        allPath = repmat(str0,length(finalCorrectPath),1);
        for i1=1:length(allPath)
            s1 = finalCorrectPath{i1};
            t1 = finalTimedPath{i1};
            % mzFrom
            allPath(i1).mzFrom = mzName{s1(1)};
            % choFrom
            allPath(i1).choFrom = dataAll{s1(1),4};
            % mzTo
            allPath(i1).mzTo = mzName{s1(end)};
            % choTo
            allPath(i1).choTo = dataAll{s1(end),4};
            % rules
            ruleS = cell(1,length(s1)-1);
            rulePath = zeros(1,length(s1)-1);
            for j1=1:length(s1)-1
                ruleS{j1} = ruleList{mmNetPattern(s1(j1),s1(j1+1))};
                rulePath(j1) = mmNetPattern(s1(j1),s1(j1+1));
            end
            allPath(i1).rules = ruleS;
            allPath(i1).rulePath = rulePath;
            % dispPath
            dispPath = [];
            for j1=1:length(s1)-1
                dispPath = [dispPath, mzName{s1(j1)},' -> '];
            end
            dispPath = [dispPath,mzName{s1(end)}];
            allPath(i1).dispPath = dispPath;
            % mzPath
            allPath(i1).mzPath = mzName(s1)';
            % choPath
            allPath(i1).choPath = dataAll(s1,4)';
            % TS
            allPath(i1).TS = t1;
        end
        
        % construct the mz_dataStruct
        mzStruct.data = data;
        mzStruct.dataAll = dataAll;
        mzStruct.finalCorrectPath = finalCorrectPath;
        mzStruct.finalTimedPath = finalTimedPath;
        mzStruct.mmNet = mmNet;
        mzStruct.mmNetS = mmNetS;
        mzStruct.mmNetPattern = mmNetPattern;
        mzStruct.mzList = mzList;
        mzStruct.mzName = mzName;
        mzStruct.ruleList = ruleList;
        mzStruct.x_final = x_final;
        mzStruct.y_final = y_final;
        mzStruct.allPath = allPath;
        mzStruct.compReal = compReal;
        save([outName,'_mzStruct.mat'], 'mzStruct');
    end

end  % end of main function

% ####################    END    ######################





